#!/usr/bin/env python

# Copyright (C) 2014 Castrense Savojardo
#     Bologna Biocomputing Group
#     University of Bologna, Italy
#     savojard@biocomp.unibo.it
#  
#  tppred2.py - This file is part of TPPRED2
#  
#  TPPRED2 is a free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published
#  by the Free Software Foundation; either version 3 of the License,
#  or (at your option) any later version.
# 
#  TPPRED2 is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with TPPRED2; if not, write to the Free Software Foundation,
#  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.


# This variable should be set to the tppred installation folder


import sys
import os

TPPRED_ROOT = os.environ.get('TPPRED_ROOT')
sys.path.append(TPPRED_ROOT)

import os

try:
    import tempfile
except ImportError:
    sys.stderr.write("Error: Python module tempfile not found.\n")
    sys.exit(1)

try:
    import subprocess
except ImportError:
    sys.stderr.write("Error: Python module subprocess not found.\n")
    sys.exit(1)

try:
    import argparse
except ImportError:
    sys.stderr.write("Error: Python module argparse not found.\n")
    sys.exit(1)

try:
    from Bio import SeqIO
except ImportError:
    sys.stderr.write("Error: Python module biopython not found.\n")
    sys.exit(1)

try:
    import numpy
except ImportError:
    sys.stderr.write("Error: Python module numpy not found.\n")
    sys.exit(1)

try:
    from scipy.stats import norm
except ImportError:
    sys.stderr.write("Error: Python module scipy not found.\n")
    sys.exit(1)

try:
    import pickle
except ImportError:
    sys.stderr.write("Error: Python module pickle not found.\n")
    sys.exit(1)

try:
    import gzip
except ImportError:
    sys.stderr.write("Error: Python module gzip not found.\n")
    sys.exit(1)

import tppred2modules.tppred2config as cfg
import tppred2modules.tppred2utils as utls

def parse_arguments():
    parser = argparse.ArgumentParser(prog = 'tppred2.py',
                                     formatter_class = argparse.RawDescriptionHelpFormatter,
                                     description = cfg.DESCRIPTION,
                                     epilog = cfg.EPILOG,
                                     usage = '%(prog)s -f FASTA_File [-o out_file]')
    options = parser.add_argument_group('OPTIONS')
    options.add_argument('-f',
                         help = 'Protein sequences in FASTA format. Required.',
                         type = argparse.FileType('r'),
                         dest = 'fasta',
                         metavar = 'FILE',
                         required = True)
    options.add_argument('-o',
                         help = 'Output prediction file. Optional, default: STDOUT.',
                         type = argparse.FileType('w'),
                         dest = 'outFile',
                         default = sys.stdout,
                         metavar = 'FILE',
                         required = False)
    ns = parser.parse_args()
    return ns

def main():
    args = parse_arguments()
    crf_model = os.path.join(TPPRED_ROOT, cfg.CRF_MODEL_FILE)
    svm_model = os.path.join(TPPRED_ROOT, cfg.SVM_MODEL_FILE)
    
    args.outFile.write("{0:30s} {1:12s} {2:6s}\n".format("ProteinID", "CleavageSite", "Score"))
    args.outFile.write("-" * 50)
    args.outFile.write("\n")
    
    for fasta in SeqIO.parse(args.fasta, 'fasta'):
        l = len(str(fasta.seq))
        seq = str(fasta.seq).replace("U", "C")[:min(l, 160)]
        fastarecfile = tempfile.NamedTemporaryFile(delete = False)
        fastarecfilename = fastarecfile.name
        SeqIO.write([fasta], fastarecfile, 'fasta')
        fastarecfile.close()
        m100, m160 = utls.compute_hmom(fastarecfilename)
        os.unlink(fastarecfilename)
        
        inputdat = utls.encode_protein(seq, m100, m160, cfg.KDSCALE, cfg.HWINDOW, cfg.AA_ORDER)
        crf_cleavage, label_prob, state_prob = utls.crf_predict(inputdat, crf_model)
        crf_prob = 0.0
        if crf_cleavage > 0:
            crf_prob = numpy.mean(label_prob[:crf_cleavage])
        else:
            crf_prob = numpy.mean(1.0 - label_prob[:min(30, len(seq))])
        svm_dat = utls.svm_encode_protein(seq, state_prob, label_prob, cfg.SMOOTHWINDOW, cfg.REGEXW)
        svm_cleavage, svm_prob = utls.svm_predict(svm_dat, svm_model)
        cleavage = "-"
        prob = 0.0
        localization = "-"
        if crf_cleavage > 0:
            localization = "M"
            if svm_cleavage > 0:
                cleavage, prob = str(svm_cleavage), svm_prob
            else:
                cleavage, prob = str(crf_cleavage), crf_prob
        else:
            prob = crf_prob
        args.outFile.write("{0:30s} {1:12s} {2:1.4f}\n".format(fasta.id[:30], cleavage, round(prob, 4)))
    args.outFile.write("-" * 50)
    args.outFile.write("\n")
    
    return 0

if __name__ == "__main__":
    r = main()
    sys.exit(r)             
    
