# Copyright (C) 2014 Castrense Savojardo
#     Bologna Biocomputing Group
#     University of Bologna, Italy
#     savojard@biocomp.unibo.it
#  
#  utils.py - This file is part of TPPRED2
#  
#  TPPRED2 is a free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published
#  by the Free Software Foundation; either version 3 of the License,
#  or (at your option) any later version.
# 
#  TPPRED2 is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with TPPRED2; if not, write to the Free Software Foundation,
#  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.

__all__ = ['compute_hmom',
           'encode_protein',
           'svm_encode_protein',
           'crf_predict',
           'svm_predict']

import os
import tempfile
import subprocess as sp

import numpy
from scipy.stats import norm

import re

import pickle
import gzip

import biocompy.crf as crf

def find_overlapping(cre, s):
    l = len(s)
    matches = []
    i = 0
    while i < l:
        mobj = cre.search(s[i:])
        if mobj != None:
            matches.append((mobj.start()+i,mobj.end()+i))
        else:
            break
        i = mobj.start()+i+1
    return matches

def compute_hmom(fastafile):
    m100, m160 = [], []
    outfile = tempfile.NamedTemporaryFile(delete = False)
    hmomfilename = outfile.name
    sp.check_call(['hmoment',fastafile,'-double','-outfile', hmomfilename], stdout = open("/dev/null", 'w'), stderr = open("/dev/null", 'w'))
    outfile.close()
    for line in open(hmomfilename).readlines()[4:]:
        line = line.split()
        m100.append(float(line[1]))
        m160.append(float(line[2]))
    os.unlink(hmomfilename)
    return m100, m160


def encode_protein(sequence, m100, m160, hscale, window, alphabet):
    dat = []
    pos = 'RK'
    neg = 'ED'
    for i in range(len(sequence)):
        h = p = n = 0.0
        subseq = sequence[max(0, i-window/2):min(len(sequence),i+window/2 + 1)]
        for aa in subseq:
            h += hscale[aa]
            p += float(aa in pos)
            n += float(aa in neg)
        h = h / len(subseq)
        p = p / len(subseq)
        n = n / len(subseq)
        v = [0.0] * len(alphabet)
        try:
            v[alphabet.index(sequence[i])] = 1.0
        except ValueError:
            pass
        if i >= 5 and i < len(sequence) - 5:
            v.extend([h, p, n, m100[i-5], m160[i-5]])
        else:
            v.extend([h, p, n, 0.0, 0.0])
        dat.append(v)
    return numpy.array(dat)

def svm_encode_protein(sequence, pstate, plabel, smooth_window, regexw):
    svmdat = []
    pstate_sm = numpy.zeros_like(pstate)
    plabel_sm = numpy.zeros_like(plabel)
    for i in range(len(sequence)):
        w = pstate[max(0, i-smooth_window):min(len(sequence),i+smooth_window+1)]
        pstate_sm[i] = numpy.mean(w)
        w = plabel[max(0, i-smooth_window):min(len(sequence),i+smooth_window+1)]
        plabel_sm[i] = numpy.mean(w)
    r2 = re.compile('R[A-Z][A-Z]')
    r3a = re.compile('R[A-Z][Y|F|L][S|A]')
    r3x = re.compile('R[A-Z][Y|F|L][A-Z]')
    r10 = re.compile('R[A-Z][F|L|I][A-Z]{2}[T|S|G][A-Z]{5}')
    rn = re.compile('[A-Z][A-Z]S')
    
    matches = {}
    
    matches['R2'] = [m[0] + 1 for m in find_overlapping(r2, sequence)]
    matches['R3a'] = [m[0] + 2 for m in find_overlapping(r3a, sequence)]
    matches['R3x'] = [m[0] + 2 for m in find_overlapping(r3x, sequence)]
    matches['R10'] = [m[0] + 9 for m in find_overlapping(r10, sequence)]
    matches['RNone'] = [m[0] for m in find_overlapping(rn, sequence)]
    
    for i in range(len(sequence)):
        regex_score = 0.0
        for r in ['R2', 'R3a', 'R3x', 'R10', 'RNone']:
            regex_score += (regexw[r] + norm.pdf(i, loc=35, scale=16)) * int(i in matches[r])
        svmdat.append([regex_score, pstate_sm[i], plabel_sm[i]])
    return numpy.array(svmdat)

def crf_predict(inputdat, crfmodel):
    crfm = crf.CRF()
    crfm.parse(crfmodel)
    target = "".join(crfm.predict([(inputdat, None)], algo = 'posterior-viterbi-sum')[0])
    label_prob = []
    state_prob = []
    for i in range(inputdat.shape[0]):
        label_prob.append(crfm.prob('P', i, label = True))
        state_prob.append(crfm.prob('F3', i, label = False))
    m = re.match('P+', target)
    if not m == None:
        cleavage = m.end()
    else:
        cleavage = 0
    return cleavage, numpy.array(label_prob), numpy.array(state_prob)

def svm_predict(svmdat, svmmodel):
    clf = pickle.load(gzip.open(svmmodel))
    pclassidx = list(clf.classes_).index(1.0)
    nclassidx = list(clf.classes_).index(-1.0)
    pred = clf.predict(svmdat)
    pred_prob = clf.predict_proba(svmdat)
    n = pred.shape[0]
    positives = []
    negatives = []
    for i in range(n):
        if pred[i] == 1.0:
            positives.append((pred_prob[i][pclassidx], i + 1))
        else:
            negatives.append((pred_prob[i][nclassidx], i + 1))
    if len(positives) > 0:
        positives.sort()
        return positives[-1][1], positives[-1][0]
    else:
        negatives.sort()
        return 0, negatives[0][0]            

    
    
